/***************************************************************************
 *                                                                         *
 *   copyright        : (C) 2011 Commissariat a l'Energie Atomique (CEA)   *
 *   author           : Gilles Mouchard                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License version 2        *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "mmu.h"

#define MMUCR "0x3b2"
#define SRR0 "0x1a"
#define SRR1 "0x1b"

#define MAX_PAGES 64

static unsigned int tlb_index = 1; // Note: entry 0 is already used for code

uint32_t get_page_size(uint32_t length)
{
	if(length <= (1 * 1024)) return 0;
	if(length <= (4 * 1024)) return 1;
	if(length <= (16 * 1024)) return 2;
	if(length <= (64 * 1024)) return 3;
	if(length <= (256 * 1024)) return 4;
	if(length <= (1024 * 1024)) return 5;
	if(length <= (16 * 1024 * 1024)) return 7;
	if(length <= (256 * 1024 * 1024)) return 9;
}

uint32_t get_page_byte_length(uint32_t page_size)
{
	return 1024 << (2 * page_size);
}

uint32_t align_address_on_page_boundary(uint32_t addr, uint32_t page_size)
{
	return addr & ~((1024 << (2 * page_size)) - 1);
}

void mmu_init()
{
	tlb_index = 1; // Note: entry 0 is already used for code
}

void load_tlb(page_t *page)
{
	if(tlb_index >= MAX_PAGES) return;
	
	uint32_t tlb_w0;
	uint32_t tlb_w1;
	uint32_t tlb_w2;
	uint32_t mmucr;
	uint32_t msr;
	
	// compute UTLB entry value
	tlb_w0 = (page->base_effective_addr & 0xfffffc00UL) | ((uint32_t) page->valid << 9) | (((uint32_t) page->ts & 0x1) << 8) | (((uint32_t) page->size & 0xfUL) << 4);
	tlb_w1 = (page->base_physical_addr & 0xfffffc00UL) | ((page->base_physical_addr >> 32) & 0xf);
	tlb_w2 = (uint32_t) page->attr & 0xffbf;
	
	// save MSR
	__asm__ volatile("mfmsr %0" : "=r" (msr));

	// disable maskable interrupts (MSR[E]=0) while updating TLB
	__asm__ volatile("wrteei 0");

	// read MMUCR
	__asm__ volatile("mfspr %0, " MMUCR "\n" : "=r" (mmucr));
	
	// write tid in MMUCR[STID]
	mmucr = (mmucr & 0xffffff00UL) | page->tid;
	__asm__ volatile("mtspr " MMUCR ", %0\n" : : "r" (mmucr));

	// load TLB
	__asm__ volatile(
	// assembly
	"tlbwe %1, %0, 0\n"      /* Set EPN, V, TS, SIZE, TID of UTLB entry word #0 */
	"tlbwe %2, %0, 1\n"      /* Set RPN, ERPN of UTLB entry word #1 */
	"tlbwe %3, %0, 2\n"      /* Set U0 U1 U2 U3 W I M G E UX UW UR SX SW SR of UTLB entry word #2 */
	// output
	: 
	// input
	: "r" (tlb_index)    // operand 0
	, "r" (tlb_w0)       // operand 1
	, "r" (tlb_w1)       // operand 2
	, "r" (tlb_w2));     // operand 3

	// invalidate shadow TLBs (context synchronization) and restore MSR
	__asm__ volatile(
	// assembly
	"isync\n"
	"mtmsr %0\n"
	// output
	:
	// input
	: "r" (msr));
	
	tlb_index++;
}
